__all__ = ('__title__',
           '__summary__',
           '__uri__',
           '__version__',
           '__author__',
           '__email__',
           '__license__',
           '__copyright__')

__title__ = 'pybnb'
__summary__ = 'A parallel branch-and-bound engine for Python'
__uri__ = 'https://github.com/ghackebeil/pybnb'
__version__ = '0.5.0'
__author__ = 'Gabriel A. Hackebeil'
__email__ = 'gabe.hackebeil@gmail.com'
__license__ = 'MIT'
__copyright__ = 'Copyright {0}'.format(__author__)
